#Licensed Materials - Property of IBM 
#Rational ClearQuest 
#Copyright IBM Corp. 2008 All Rights Reserved 

package cqcli::findrecord;

use strict;
use warnings;
require Exporter;

use Error qw(:try);
use cqcli::Exception;
use Inline::Java qw(caught cast) ;
use cqcli::Util;
use cqcli::IOHandler;
use cqcli::commonservice;
use Text::ParseWords;
use Getopt::Long 2.34;


our @ISA = qw(Exporter);

our %EXPORT_TAGS = ( 'all' => [ qw() ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(findrecord);

our $VERSION = '0.01';

#######################################################################
our @ARGVBAACKUP = ();
our %RepetitionCount = ();
our $isLoginOpt = undef;
our $CliSession = undef;
our $JavaReqObj = undef;
our %CmdInfo = ( 'cmdstring' => undef, 'cmdname'  => "find_record");
our %CmdLogin = ('database'	=>	undef,'schema'=>  undef,
			 'user'		=>	undef,'password'=>  undef);
our $BLANK_MARKER_SET = 0;
our	%DbIdDef = ();
our	%RIdDef = ();;
our %FieldDef = ();
our %remove=();
our %DisplayConfiguration = ();
our %force=();

our $validateFieldNames = 0;

our	$TypeDef = "";

#########################################################################

#########################################################################

#########handle getopt warnings begin########################

sub InitParser
{
	my @tmpARGV = @_;
	@ARGVBAACKUP = @ARGV;
	@ARGV = undef;
	$isLoginOpt = undef;
	%RepetitionCount = ();
	%CmdInfo = ('cmdstring' => undef, 'cmdname'  => "find_record");
	%CmdLogin =('database'	=>undef,'schema'=>  undef,
				'user'	=>	undef,'password'=>  undef);
	
	our %DbIdDef = ();
	our %RIdDef = ();;
	our %FieldDef = ();
	our %DisplayConfiguration = ();
	our %remove=();
	our %force=();
	$validateFieldNames = 0;
	
	$TypeDef = "";
	
	$CmdInfo{'cmdstring'} =[@tmpARGV];
	($CmdInfo{'cmdname'}  , @ARGV ) = @{$CmdInfo{'cmdstring'}};
	$CliSession = GetSession();
	$JavaReqObj = GetJavaHash();
	return 1;
}

sub CleanUpParser()
{

	@ARGV = @ARGVBAACKUP;
	%RepetitionCount = ();
	%CmdInfo = ('cmdstring' => undef, 'cmdname'  => "find_record");
	%CmdLogin =('database'	=>undef,'schema'=>  undef,
				'user'	=>	undef,'password'=>  undef);
	
	our	%DbIdDef = ();
	our	%RIdDef = ();;
	our %FieldDef = ();
	our %DisplayConfiguration = ();
	$validateFieldNames = 0;
	our %remove=();
	our %force=();
	our	$TypeDef = "";
	
	$CmdInfo{'cmdstring'} = undef;
	$CmdInfo{'cmdname'} =undef;
	$CliSession = undef;
	$JavaReqObj = undef;
}

#########handle getopt warnings end#####################################



sub handle_help_usage
{
	my $opt = $_[0];
	
	if($opt eq "help")	{
		ThrowUsageException($CmdInfo{'cmdname'},$opt);
	}
	else{
		ThrowUsageException($CmdInfo{'cmdname'},$opt);
	}
	return 1;
}

sub handle_ask_mode
{
	my $mode = shift;
	SetCQCLIAskMode(1);
	return 1;
}

sub handle_login_data
{
	my ($opt , $val) = @_;
	my $msg = undef;
	$isLoginOpt = 1;

	if($RepetitionCount{$opt}) {
		ThrowSyntaxErrorException("CQCLI_Repeat_Option","-$opt", $CmdInfo{'cmdname'} );
	}else {
		$RepetitionCount{$opt} = 1;
	}

	if( ($_[0] eq "database") && ($val)	){
		$CmdLogin{'database'} = $val;
	}elsif(($_[0] eq "user") && ($val)) {
		$CmdLogin{'user'} = $val;
	}elsif(($_[0] eq "password") && ($val)) {
		$CmdLogin{'password'} = $val;
	}elsif(($_[0] eq "password") && ($val eq "")){
		$CmdLogin{'password'} = "";
		$BLANK_MARKER_SET = 1 ;
	}
	elsif((($_[0] eq "schema") || ($_[0] eq "master")) && ($val) ) {
		$CmdLogin{'schema'} = $val; 
	}else{
		ThrowSyntaxErrorException("CQCLI_Missing_Value","-$opt",$CmdInfo{'cmdname'});
	}
	return 1;
}

sub handle_recordid_list
{
	my ($opt , $val) = @_;
	my $recordid = undef;
	
	my @rid_type = undef;
	my ($rid,$rtype) = (undef,undef);
	
	if( ($opt eq "record") && ($val)	){
		$validateFieldNames = 0;
		@rid_type = split (' ',$val);
		$rtype= $rid_type[0];$rid= $rid_type[1]; 
		GetRecordTypeInfoObj($rtype);
		$RIdDef{$rid}=$rtype;
	}elsif(($opt eq "dbid") && ($val)) {
		$DbIdDef{$val}=$TypeDef
	}elsif(($opt eq "input") && ($val)){
		if (-e $val){
			open(RIN, "<$val")|| ThrowException('CQCLI_Failure_FileOpen',$val);
			while (<RIN>) {
				$recordid = $_;
				chomp($recordid);
				$recordid = cqclitrim($recordid);
				$RIdDef{$recordid}=$TypeDef;
			}
			close(RIN)|| CLIPrintIdOutput('CQCLI_Info_FileClose',$val);
		}else {
			ThrowException('CQCLI_Failure_FileOpen',$val);
		}

	}else{
		ThrowSyntaxErrorException("CQCLI_Missing_Value","-$opt",$CmdInfo{'cmdname'});
	}
	return 1;

}

sub handle_recordtype_list
{
	my ($opt , $type) = @_;
	my $msg = undef;
	if($RepetitionCount{'$opt'} ) {
		ThrowSyntaxErrorException("CQCLI_Repeat_Option","-$opt", $CmdInfo{'cmdname'} );
	}else {
		$RepetitionCount{'$opt'} = 1;
	}

	if($type){
		$TypeDef = $type;
	}else {
		ThrowSyntaxErrorException("CQCLI_Missing_Value","-$opt", $CmdInfo{'cmdname'} );
	}
	return 1;
}

sub handle_display_options
{
	my ($opt) = @_;
	
	#all_fields|af|sort_names|sn|type_names|tn|value|va|no_recname|nr|no_headers|nh|no_names|nn|no_values|nv|no_dup_info|nd
	if($opt eq "all_fields" || $opt eq "af" ) {
		$DisplayConfiguration{'ALL_FIELDS'} = 1;
	}
	elsif($opt eq "sort_names" || $opt eq "sn" ) {
		$DisplayConfiguration{'SORT_NAMES'} = 1;
	}
	elsif($opt eq "type_names" || $opt eq "tn" ) {
		$DisplayConfiguration{'TYPE_NAMES'} = 1;
	}
	elsif($opt eq "value" || $opt eq "va" ) {
		$DisplayConfiguration{'VALUE'} = 1;
	}
	elsif($opt eq "no_recname" || $opt eq "nr" ) {
		$DisplayConfiguration{'NO_RECORD_HEADER'} = 1;
	}
	elsif($opt eq "no_headers" || $opt eq "nh" ) {
		$DisplayConfiguration{'NO_HEADER'} = 1;
	}
	elsif($opt eq "no_names" || $opt eq "nn" ) {
		$DisplayConfiguration{'NO_NAMES'} = 1;
	}
	elsif($opt eq "no_values" || $opt eq "nv" ) {
		$DisplayConfiguration{'NO_VALUES'} = 1;
	}
	elsif($opt eq "no_dup_info" || $opt eq "nd" ) {
		$DisplayConfiguration{'NO_DUP_INFO'} = 1;
	}
	else {
		ThrowSyntaxErrorException("CQCLI_Missing_Value","Display field config error", $CmdInfo{'cmdname'} );
	}
	
	return 1;
}

sub handle_field_list
{
	my ($cnt , @tmpARGV) = @_;
	my $msg = undef;

	if($tmpARGV[$cnt+1])
	{
		$FieldDef{$tmpARGV[$cnt+1]} = 1;
		splice @ARGV,0,2;
	}else {
		ThrowSyntaxErrorException('CQCLI_Missing_FName',$tmpARGV[$cnt],$CmdInfo{'cmdname'});
	}

	return 1;
}


sub handle_exclusion_fields
{
	my ($opt , $field) = @_;
	my $msg = undef;

	if($field){
		$remove{$field} = 1;
	}else {
		ThrowSyntaxErrorException("CQCLI_Missing_Value","-$opt", $CmdInfo{'cmdname'} );
	}
	return 1;
}


sub parseForHelpUsage()
{
	my $ret = undef;

	my $helpRules = {'help|h|?'		=>	\&handle_help_usage	,
					 'man'			=>	\&handle_help_usage	};
	
	Getopt::Long::Configure('autoabbrev','ignore_case','permute','pass_through');
	$ret = GetOptions( %{$helpRules} );
	Getopt::Long::Configure('default');

	return 1;
}

sub parseForInterActiveOption 
{
	my $ret = undef;
	my $askRules = {'ask!'	=>	\&handle_ask_mode	};
	Getopt::Long::Configure('posix_default','ignore_case','permute','pass_through');
	$ret = GetOptions( %{$askRules} );
	Getopt::Long::Configure('default');

	return 1;
}

sub parseForUnknownOptions
{
	my @tmpARGV = @ARGV;
	my $lmt = $#tmpARGV;
	my $msg = undef;
	
	
	for (my $cnt = 0; $cnt <= $lmt;$cnt++)  
	{	
		if( $tmpARGV[$cnt] =~ /^-(f(ield)?$|fi(eld)?$|fie(ld)?$|fiel(d)?$|field?$)/i )
		{
		
			handle_field_list($cnt,@tmpARGV);
			$cnt=$cnt+1;
			next;
		}
		elsif($tmpARGV[$cnt] =~ /^(-)(.+)/)
		{
			my $fname = $2;

			$FieldDef{$fname}=1;	
			splice @ARGV,0,2;
			next;
		}
		elsif($tmpARGV[$cnt] =~ /^(\+)(.+)/)
		{
			my $fname = $2;

			$force{$fname}=1;
			$FieldDef{$fname}=1;	
			splice @ARGV,0,2;
			next;
		}
		elsif($tmpARGV[$cnt] =~ /^(\^)(.+)/)
		{
			my $fname = $2;
			$remove{$fname}=1;	
			splice @ARGV,0,2;
			next;
		}
		elsif($tmpARGV[$cnt] =~ /^(-+)/){
			ThrowSyntaxErrorException('CQCLI_Invalid_Option',$tmpARGV[$cnt]);
		}else{
			$RIdDef{$tmpARGV[$cnt]} = $TypeDef;
			shift @ARGV;
			next;
		}
		
	}
	return 1;
}


sub parseForLoginCredentials 
{
	my $ret = undef;
	my $CommonRules = {
		'database|dbs|d:s'					=>  \&handle_login_data			,
		'user|login|u|l:s'					=>	\&handle_login_data			,
		'schema|s|m|master:s'				=>	\&handle_login_data			,
		'password|p:s'						=>	\&handle_login_data			,
	};
	Getopt::Long::Configure( 'autoabbrev','ignore_case','permute','pass_through');
	$ret = GetOptions( %{$CommonRules} );
	Getopt::Long::Configure('default');
	return 1;
}

sub parseForCoreOptions()
{
	my $ret = undef;
	my $CoreOptRules = {
				'type:s'   =>	\&handle_recordtype_list , 
				'input:s'  =>   \&handle_recordid_list ,
				'record:s'  =>	\&handle_recordid_list ,
				'dbid:s'   =>	\&handle_recordid_list,
				'all_fields|af' => \&handle_display_options,
				'sort_names|sn' => \&handle_display_options,
				'type_names|tn' => \&handle_display_options,
				'value|va'      => \&handle_display_options,
				'no_recname|nr' => \&handle_display_options,
				'no_headers|nh' => \&handle_display_options,
				'no_names|nn'   => \&handle_display_options,
				'no_values|nv'  => \&handle_display_options,
				'no_dup_info|nd'=> \&handle_display_options,
				'exclude:s' => \&handle_exclusion_fields
	};
	Getopt::Long::Configure( 'autoabbrev','ignore_case','permute','pass_through');
	$ret = GetOptions( %{$CoreOptRules} );
	Getopt::Long::Configure('default');

	return 1;
}

##################################################################################################

sub ValidateSessionInfo()
{
	
	if($isLoginOpt){
		ValidateLoginCredentials();
		if($TypeDef eq "") {
			$TypeDef = getDefaultRecordtype();
		}
		return;
	}elsif($CliSession){

		if($TypeDef eq "") {
			$TypeDef = getDefaultRecordtype();
		}
		return;
	}else{
		ValidateLoginCredentials();
		if($TypeDef eq "") {
			$TypeDef = getDefaultRecordtype();
		}
		return;
	}

}
sub ValidateLoginCredentials()
{
	if(!($CmdLogin{'user'})){
		$CmdLogin{'user'} = GetUserName();
	}
	if(!($CmdLogin{'password'}) && $BLANK_MARKER_SET == 0){
		$CmdLogin{'password'} = GetPassword();
	}
	if(!($CmdLogin{'database'})){
		$CmdLogin{'database'} = GetDBName();
	}
	if(!($CmdLogin{'schema'})){
		$CmdLogin{'schema'} = GetSchemaName();
	}
	$CliSession=CreateSession(%CmdLogin);
	# is not a command scope session
	SetSessionScope(0);
}

sub ValidateFieldNames()
{
	my @fieldnames = keys %FieldDef;

	my $recordtypeobj = GetRecordTypeInfoObj($TypeDef);
	my $expectedfield = undef ;
	
	foreach my $fld (@fieldnames)
	{
		$expectedfield = $recordtypeobj->isValidFieldName($fld);

		if($expectedfield) {
			if($expectedfield =~ /\,/ ){
				ThrowValidationException('CQCLI_Failure_FieldAmbiguity',$fld,$expectedfield,$TypeDef);
			}
			else {
				delete $FieldDef{$fld};
				$FieldDef{$expectedfield}=1;
			}
		}
		else {
				ThrowValidationException("CQCLI_Invalid_FieldInput",$TypeDef);
		}

	}	

}

sub ValidateRemoveFieldNames()
{
	my @fieldnames = keys %remove;

	my $recordtypeobj = GetRecordTypeInfoObj($TypeDef);
	my $expectedfield = undef ;
	
	foreach my $fld (@fieldnames)
	{
		$expectedfield = $recordtypeobj->isValidFieldName($fld);

		if($expectedfield) {
			if($expectedfield =~ /\,/ ){
				ThrowValidationException('CQCLI_Failure_FieldAmbiguity',$fld,$expectedfield,$TypeDef);
			}
			else {
				delete $remove{$fld};
				$remove{$expectedfield}=1;
			}
		}
		else {
				ThrowValidationException("CQCLI_Invalid_FieldInput",$TypeDef);
		}

	}	

}

sub ValidateForceFieldNames()
{
	my @fieldnames = keys %force;

	my $recordtypeobj = GetRecordTypeInfoObj($TypeDef);
	my $expectedfield = undef ;
	
	foreach my $fld (@fieldnames)
	{
		$expectedfield = $recordtypeobj->isValidFieldName($fld);

		if($expectedfield) {
			if($expectedfield =~ /\,/ ){
				ThrowValidationException('CQCLI_Failure_FieldAmbiguity',$fld,$expectedfield,$TypeDef);
			}
			else {
				delete $force{$fld};
				$force{$expectedfield}=1;
			}
		}
		else {
				ThrowValidationException("CQCLI_Invalid_FieldInput",$TypeDef);
		}

	}	

}

sub ValidateForceFields()
{

 	my ($keys1 , $keys2) = (undef,undef);
	my $id = undef;
	$keys1 = keys %force;
	$keys2 = keys %remove;
	my $ffields = undef; 

	if( ($keys1 >0) && ($keys2 >0 )){
		foreach $ffields (keys %force){
			if($remove{$ffields}){
				delete $remove{$ffields};
			}
		}
		
	}

}


sub ValidateRecordIds()
{
	 
	my ($keys1 , $keys2) = (undef,undef);
	my $id = undef;
	$keys1 = keys %RIdDef;
	$keys2 = keys %DbIdDef;
	
	if( ($keys1 <=0) && ($keys2 <=0 ) && IsAPIMode()) {
		ThrowValidationException('CQCLI_Failure_MissingIds');
	}
	elsif( ($keys1 <=0) && ($keys2 <=0 ))
	{
		if(!CheckPromptModeValidity('CQCLI_Missing_IdsWarn') && AskForYN('CQCLI_Prompt_AskRid'))
		{
			do{
				$id = GetRecordId();
				$RIdDef{$id} = $TypeDef;
			} while(AskForYN('CQCLI_Prompt_Continue_Record')) ;
			
		}
		if(!CheckPromptModeValidity('CQCLI_Missing_IdsWarn') && AskForYN('CQCLI_Prompt_AskDbid'))
		{
			do{
				$id = GetDbId();
				$DbIdDef{$id} = $TypeDef;
			} while(AskForYN('CQCLI_Prompt_Continue_Record')) ;
		}
		$keys1 = keys %RIdDef;
		$keys2 = keys %DbIdDef;
		if( ($keys1 <=0) && ($keys2 <=0 )) {
			ThrowValidationException('CQCLI_Failure_MissingIds');
		}
	}
}



sub SetFieldNames()
{
	#set request object for successfull field validation
	my $fieldhash = GetJavaHash();
	my $fld = undef;
	
	my $removekeys = undef;
	my $removehash = GetJavaHash();

	my $keys1 = keys %remove;
	
	foreach $removekeys (keys %remove)
	{	
		if($FieldDef{$removekeys}) {
			delete $FieldDef{$removekeys};
		}
	}
	
	foreach $fld (keys %FieldDef) {
		$fieldhash->put(lc($fld),$FieldDef{$fld});
	}
	
	foreach $fld (keys %remove) {
		$removehash->put(lc($fld),$remove{$fld});
	}
	
	$JavaReqObj->put("field",$fieldhash);

	if($keys1 > 0) {
		$JavaReqObj->put("remove",$removehash);
	}
}

sub SetDisplayConfig()
{
	#set request object for successfull field validation
	my $displayhash = GetJavaHash();
	my $fld = undef;
	foreach $fld (keys %DisplayConfiguration) {
		$displayhash->put($fld,$DisplayConfiguration{$fld})
	}
	$JavaReqObj->put("display",$displayhash);
}

sub SetRecordIds
{
	my ($rectypeinfo,$key,$flag,$ridhash,$dbidhash) = (undef,undef,undef,undef,undef);
	try {
		$rectypeinfo = GetRecordInfoObj();
		if(keys %RIdDef) {
			$ridhash = GetJavaHash();
			foreach $key (keys %RIdDef) {
				if($RIdDef{$key} eq ""){
					$RIdDef{$key} = $TypeDef;
				}
				$ridhash->put($key,$RIdDef{$key});
			}
			$flag = 1;
			$JavaReqObj->put("record",$ridhash);
		}		
		if(keys %DbIdDef){
			$dbidhash = GetJavaHash(); 
			foreach $key (keys %DbIdDef) {
				if($DbIdDef{$key} eq ""){
					$DbIdDef{$key} = $TypeDef;
				}
				$dbidhash->put($key,$DbIdDef{$key});
			}
			$flag = 0;
			$JavaReqObj->put("dbid",$dbidhash);
		}
	}
	catch cqcli::Exception with {
		my $ex = shift;	
		throw $ex;
	}
	otherwise {
		if (caught("java.lang.Exception")){
			ThrowJavaException($@,"CQCLI_RecordNotFound");
		}
		else{
			ThrowException("CQCLI_Failure_Unexpected",$@);
		}
	}finally {};
}



###########################################################################################


sub Parse()
{
	try {
		parseForHelpUsage();
		parseForInterActiveOption();
		parseForLoginCredentials();
		parseForCoreOptions();
		parseForUnknownOptions();
	}
	catch cqcli::Exception with {
		my $ex = shift;		
		throw $ex;
	}
	otherwise {
		if (caught("java.lang.Exception")){
			ThrowJavaException($@,"CQCLI_ErrJava_ParserUnknown");
		}
		else{
			ThrowParseException('CQCLI_Err_ParserUnknown',$@);
		}
		
    }
	finally {
		Getopt::Long::Configure('default');
	};

	return 1;
}

sub Validate()
{
	try {
		ValidateSessionInfo();
		
		if($validateFieldNames == 0) {
			ValidateFieldNames();
		}
		
		my $keys1 = keys %remove;

		if($keys1 >0){
			ValidateRemoveFieldNames();
		}
		
		my $keys2 = keys %force;

		if($keys2 >0){
			ValidateForceFieldNames();
		}

		ValidateForceFields();
		ValidateRecordIds();
	}
	catch cqcli::Exception with {
		my $ex = shift;		
		throw $ex;
	}
	otherwise {
		if (caught("java.lang.Exception")){
			ThrowJavaException($@,"CQCLI_ErrJava_ValidationUnknown");
		}
		else{
			ThrowValidationException('CQCLI_Err_ValidationUnknown',$@);
		}		
    }
	finally {	};

	return 1;
}

sub SetObjects()
{
	SetFieldNames();
	SetRecordIds();
	SetDisplayConfig();
}

sub findrecord
{
	my @tmpCmd = @_;
	my $req = undef;
	try {
		InitParser(@tmpCmd);
		Parse();
		Validate();
		SetObjects();
		$req =  GetJavaPayload("find_record",$JavaReqObj);
		SetRequestObject($req);
		return 1;
	}finally {
		CleanUpParser();
	};
}

1;
